// xkeymacsDlg.cpp : implementation file
//

#include "stdafx.h"
#include "xkeymacs.h"
#include "xkeymacsDlg.h"
#include "MainFrm.h"
#include <Imm.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

enum {INITIAL_SIZE	= 51200};
enum {EXTEND_SIZE	= 25600};

/////////////////////////////////////////////////////////////////////////////
// CXkeymacsDlg dialog


CXkeymacsDlg::CXkeymacsDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CXkeymacsDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CXkeymacsDlg)
	m_nKillRingMax = 0;
	//}}AFX_DATA_INIT
}


void CXkeymacsDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CXkeymacsDlg)
	DDX_Control(pDX, IDC_USE_DEFAULT, m_cUseDefault);
	DDX_Control(pDX, IDC_APPLICATION, m_cApplication);
	DDX_Text(pDX, IDC_KILL_RING_MAX, m_nKillRingMax);
	DDV_MinMaxInt(pDX, m_nKillRingMax, 0, 2147483647);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CXkeymacsDlg, CDialog)
	//{{AFX_MSG_MAP(CXkeymacsDlg)
	ON_BN_CLICKED(IDC_APPLY, OnApply)
	ON_CBN_SELCHANGE(IDC_APPLICATION, OnSelchangeApplication)
	ON_BN_CLICKED(IDC_CLEAR_ALL, OnClearAll)
	ON_BN_CLICKED(IDC_SELECT_ALL, OnSelectAll)
	ON_BN_CLICKED(IDC_LOAD_DEFAULT, OnLoadDefault)
	ON_BN_CLICKED(IDC_USE_DEFAULT, OnUseDefault)
	ON_CBN_DROPDOWN(IDC_APPLICATION, OnDropdownApplication)
	ON_BN_CLICKED(IDC_CO, OnCo)
	ON_BN_CLICKED(IDC_CO2, OnCo2)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CXkeymacsDlg message handlers

BOOL CXkeymacsDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here
	SetForegroundWindow();

	InitApplicationList();

	LoadRegistryData();
	m_szCurrentApplication.LoadString(IDS_DEFAULT);
	SetDialogData(m_szCurrentApplication);
	EnableUseDefaultButton(FALSE);
	GetDlgItem(IDC_LOAD_DEFAULT)->EnableWindow(FALSE);


	OSVERSIONINFO verInfo = {0};
	verInfo.dwOSVersionInfoSize = sizeof (verInfo);
	GetVersionEx(&verInfo);

	switch (verInfo.dwPlatformId) {
	case VER_PLATFORM_WIN32_NT:
		GetDlgItem(IDC_LEFT_ALT)->ShowWindow(SW_SHOWNA);
		GetDlgItem(IDC_RIGHT_ALT)->ShowWindow(SW_SHOWNA);
		GetDlgItem(IDC_ALT)->ShowWindow(SW_HIDE);
		break;

	case VER_PLATFORM_WIN32_WINDOWS:
		GetDlgItem(IDC_LEFT_ALT)->ShowWindow(SW_HIDE);
		GetDlgItem(IDC_RIGHT_ALT)->ShowWindow(SW_HIDE);
		GetDlgItem(IDC_ALT)->ShowWindow(SW_SHOWNA);
		break;

	default:
		AfxMessageBox(IDS_ERR_WINDOWS_VERSION);
		return 0;
	}

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

// This method initializes data in the registry, or retrieves and validates registry data.
// bSaveAndValidate specifies a flag that indicates 
// whether registry data is being initialized (FALSE) or data is being retrieved (TRUE). 
void CXkeymacsDlg::UpdateRegistryData(BOOL bSaveAndValidate)
{
	CString		szEntry;
	CString		szApplicationName;
	CString		szApplicationTitle;

	for (int nApp = 0; nApp < MAX_APP; nApp++) {
		// application name
		CString szSection;
		szSection.LoadString(IDS_REG_SECTION_APPLICATION);
		szEntry.Format(IDS_REG_ENTRY_APPLICATION, nApp);
		if (bSaveAndValidate) {	// retrieve
			szApplicationName = AfxGetApp()->GetProfileString(szSection, szEntry);
			if (szApplicationName.IsEmpty()) {
				if (nApp) {
					continue;
				} else {
					szApplicationName.LoadString(IDS_DEFAULT);
				}
			}
			m_XkeymacsData[nApp].ClearAll();
			m_XkeymacsData[nApp].SetApplicationName(szApplicationName);
		} else {				// initialize
			szApplicationName = m_XkeymacsData[nApp].GetApplicationName();
			if (szApplicationName.IsEmpty()) {
				continue;
			}
			AfxGetApp()->WriteProfileString(szSection, szEntry, szApplicationName);
		}

		// application title
		szEntry.LoadString(IDS_REG_ENTRY_APPLICATOIN_TITLE);
		if (bSaveAndValidate) {	// retrieve
			szApplicationTitle = AfxGetApp()->GetProfileString(szApplicationName, szEntry);
			m_XkeymacsData[nApp].SetApplicationTitle(szApplicationTitle);
		} else {				// initialize
			szApplicationTitle = m_XkeymacsData[nApp].GetApplicationTitle();
			AfxGetApp()->WriteProfileString(szApplicationName, szEntry, szApplicationTitle);
		}

		for (int nCommandType = 0; nCommandType < MAX_COMMAND_TYPE; nCommandType++) {
			for (int nKey = 0; nKey < MAX_KEY; nKey++) {
				if (!GetControlID(nCommandType, nKey)) {
					continue;
				}

				switch (nCommandType) {
				case NONE:
					szEntry.Format(IDS_REG_ENTRY_NONE, nKey);
					break;
				case CONTROL:
					szEntry.Format(IDS_REG_ENTRY_CONTROL, nKey);
					break;
				case CONTROL_SHIFT:
					szEntry.Format(IDS_REG_ENTRY_CONTROL_SHIFT, nKey);
					break;
				case CONTROLX:
					szEntry.Format(IDS_REG_ENTRY_CONTROLX, nKey);
					break;
				case CONTROL_META_SHIFT:
					szEntry.Format(IDS_REG_ENTRY_CONTROL_META_SHIFT, nKey);
					break;
				case META:
					szEntry.Format(IDS_REG_ENTRY_META, nKey);
					break;
				case META_SHIFT:
					szEntry.Format(IDS_REG_ENTRY_META_SHIFT, nKey);
					break;
				default:
					ASSERT(0);
					break;
				}

				// if nKey is back slash
				if (nKey == 0xDC || nKey == 0xE2) {
					szEntry += '\\';
				}

				int nEnable = 0;
				if (bSaveAndValidate) {	// retrieve
					nEnable = AfxGetApp()->GetProfileInt(szApplicationName, szEntry, 1);
					m_XkeymacsData[nApp].SetEnable(nCommandType, nKey, nEnable);
				} else {				// initialize
					nEnable = m_XkeymacsData[nApp].GetEnable(nCommandType, nKey);
					AfxGetApp()->WriteProfileInt(szApplicationName, szEntry, nEnable);
				}
			}
		}

		// kill-ring-max
		szEntry.LoadString(IDS_REG_ENTRY_KILL_RING_MAX);
		if (bSaveAndValidate) {	// retrieve
			int nKillRingMax = AfxGetApp()->GetProfileInt(szApplicationName, szEntry, 0);
			m_XkeymacsData[nApp].SetKillRingMax(nKillRingMax);
		} else {				// initialize
			int nKillRingMax = m_XkeymacsData[nApp].GetKillRingMax();
			AfxGetApp()->WriteProfileInt(szApplicationName, szEntry, nKillRingMax);
		}
	}
}

// load registry data
void CXkeymacsDlg::LoadRegistryData()
{
	UpdateRegistryData(TRUE);
}

// save registry data
void CXkeymacsDlg::SaveRegistryData()
{
	UpdateRegistryData(FALSE);

	SetDllData();
	CMainFrame* pMainFrame = (CMainFrame*)AfxGetMainWnd();
	pMainFrame->m_pXkeymacsDll->UpdateData();
}

void CXkeymacsDlg::OnCancel() 
{
	// TODO: Add extra cleanup here
	LoadRegistryData();
	
	CDialog::OnCancel();
}

void CXkeymacsDlg::OnApply() 
{
	// TODO: Add your control notification handler code here
	GetDialogData(m_szCurrentApplication);

	// It should be in SaveRegistry Data(); ?!
	HKEY hkey;
	CString szSubKey;
	szSubKey.LoadString(IDS_REGSUBKEY_DATA);
	VERIFY(RegOpenKeyEx(HKEY_CURRENT_USER, szSubKey, 0, KEY_ALL_ACCESS, &hkey) == ERROR_SUCCESS);
	// I am sure that I have to call DeleteAllRegistryData only one time, but...
	for (int nApp = 0; nApp < MAX_APP; nApp++) {
		DeleteAllRegistryData(hkey);
	}
	VERIFY(RegCloseKey(hkey) == ERROR_SUCCESS);

	SaveRegistryData();
}

void CXkeymacsDlg::OnOK() 
{
	// TODO: Add extra validation here
	OnApply();
	
	CDialog::OnOK();
}

// set dialog data
void CXkeymacsDlg::SetDialogData(CString szApplicationName)
{
	UpdateDialogData(szApplicationName, FALSE);
}

// get dialog Data
void CXkeymacsDlg::GetDialogData(CString szApplicationName)
{
	if (IsDlgButtonChecked(IDC_USE_DEFAULT)) {
		for (int nApp = 0; nApp < MAX_APP; nApp++) {
			if (m_XkeymacsData[nApp].GetApplicationName() == m_szCurrentApplication) {
				break;
			}
		}
		if (nApp < MAX_APP) {
			m_XkeymacsData[nApp].ClearAll();
		}
	} else {
		UpdateDialogData(szApplicationName, TRUE);
	}
}

// set/get dialog data
void CXkeymacsDlg::UpdateDialogData(CString szApplicationName, BOOL bSaveAndValidate)
{
	BOOL bCheckUseDefault = FALSE;

	int nApp;
	for (nApp = 0; nApp < MAX_APP; nApp++) {
		if (m_XkeymacsData[nApp].GetApplicationName() == szApplicationName) {
			break;
		}
	}

	if (nApp == MAX_APP) {
		if (bSaveAndValidate) {	// GetDialogData
			for (nApp = 0; nApp < MAX_APP; nApp++) {
				CString sz = m_XkeymacsData[nApp].GetApplicationName();
				if (sz.IsEmpty()) {
					m_XkeymacsData[nApp].SetApplicationName(szApplicationName);
					break;
				}
			}
			if (nApp == MAX_APP) {
				return;
			}
		} else {				// SetDialogData
			for (nApp = 0; nApp < MAX_APP; nApp++) {
				if (IsDefault(m_XkeymacsData[nApp].GetApplicationName())) {
					bCheckUseDefault = TRUE;
					break;
				}
			}
			if (nApp == MAX_APP) {
				SetAllDialogData(1);
				return;
			}
		}
	}

	// SetDialogData
	if (!bSaveAndValidate) {
		CheckUseDefault(bCheckUseDefault);
	}
	if (IsDefault(szApplicationName)) {
		EnableUseDefaultButton(FALSE);
	} else {
		EnableUseDefaultButton(TRUE);
	}

	// application title
	static CString szApplicationTitle;
	if (bSaveAndValidate) {	// GetDialogData
		if (!IsDefault(m_szCurrentApplication)) {
			m_XkeymacsData[nApp].SetApplicationTitle(szApplicationTitle);
		}
		szApplicationTitle.Empty();
	} else {				// SetDialogData
		CString szListItem;
		GetApplicationTitle(szListItem);
		int nEndTitle = szListItem.ReverseFind('(');
		if (nEndTitle > 0) {
			szApplicationTitle = szListItem.Left(nEndTitle);
		}
	}

	// on/off
	for (int nCommandType = 0; nCommandType < MAX_COMMAND_TYPE; nCommandType++) {
		for (int nKey = 0; nKey < MAX_KEY; nKey++) {
			int nControlID = GetControlID(nCommandType, nKey);
			if (!nControlID) {
				continue;
			}

			if (bSaveAndValidate) {	// GetDialogData
				m_XkeymacsData[nApp].SetEnable(nCommandType, nKey, IsDlgButtonChecked(nControlID));
			} else {				// SetDialogData
				if (m_XkeymacsData[nApp].GetEnable(nCommandType, nKey) == BST_CHECKED) {
					CheckDlgButton(nControlID, BST_CHECKED);
				} else {
					CheckDlgButton(nControlID, BST_UNCHECKED);
				}
			}
		}
	}
	// same key
	if (bSaveAndValidate) {	// GetDialogData
		if (m_XkeymacsData[nApp].GetEnable(CONTROL, 'O') == BST_UNCHECKED) {
			m_XkeymacsData[nApp].SetEnable(CONTROL, 'O', IsDlgButtonChecked(IDC_CO2) ? 2 : BST_UNCHECKED);
		}
	} else {				// SetDialogData
		if (m_XkeymacsData[nApp].GetEnable(CONTROL, 'O') == 2) {
			CheckDlgButton(IDC_CO2, BST_CHECKED);
		} else {
			CheckDlgButton(IDC_CO2, BST_UNCHECKED);
		}
	}

	// kill-ring-max
	if (bSaveAndValidate) {	// GetDialogData
		m_XkeymacsData[nApp].SetKillRingMax(GetDlgItemInt(IDC_KILL_RING_MAX));
	} else {				// SetDialogData
		SetDlgItemInt(IDC_KILL_RING_MAX, m_XkeymacsData[nApp].GetKillRingMax());
	}
}

// return control ID or 0
int CXkeymacsDlg::GetControlID(int nCommandType, int nKey)
{
	int nControlID = 0;

	switch (nCommandType) {
	case NONE:
		switch (nKey) {
		case 'U':
			nControlID = IDC_CXU;
			break;
		case VK_ESCAPE:
			nControlID = IDC_ESCAPE;
			break;
		case VK_LMENU:
			nControlID = IDC_LEFT_ALT;
			break;
		case VK_RMENU:
			nControlID = IDC_RIGHT_ALT;
			break;
		case VK_MENU:
			nControlID = IDC_ALT;
			break;
		default:
			break;
		}
		break;

	case CONTROL:
		switch (nKey) {
		case 'A':
			nControlID = IDC_CA;
			break;
		case 'B':
			nControlID = IDC_CB;
			break;
		case 'D':
			nControlID = IDC_CD;
			break;
		case 'E':
			nControlID = IDC_CE;
			break;
		case 'F':
			nControlID = IDC_CF;
			break;
		case 'G':
			nControlID = IDC_CG;
			break;
		case 'H':
			nControlID = IDC_CH;
			break;
		case 'I':
			nControlID = IDC_CI;
			break;
		case 'K':
			nControlID = IDC_CK;
			break;
		case 'L':
			nControlID = IDC_CL;
			break;
		case 'M':
			nControlID = IDC_CM;
			break;
		case 'N':
			nControlID = IDC_CN;
			break;
		case 'O':
			nControlID = IDC_CO;
			break;
		case 'P':
			nControlID = IDC_CP;
			break;
		case 'Q':
			nControlID = IDC_CQ;
			break;
		case 'R':
			nControlID = IDC_CR;
			break;
		case 'S':
			nControlID = IDC_CS;
			break;
		case 'T':
			nControlID = IDC_CT;
			break;
		case 'U':
			nControlID = IDC_CU;
			break;
		case 'V':
			nControlID = IDC_CV;
			break;
		case 'W':
			nControlID = IDC_CW;
			break;
		case 'Y':
			nControlID = IDC_CY;
			break;
		case 'Z':
			nControlID = IDC_CZ;
			break;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			nControlID = IDC_CDIGIT;
			break;
		case VK_RETURN:
			nControlID = IDC_CENTER;
			break;
		case VK_SPACE:
			nControlID = IDC_CSPACE;
			break;
		case 0xBF:	// VK_OEM_2	Used for miscellaneous characters; it can vary by keyboard. 
					//			Windows 2000/XP: For the US standard keyboard, the '/?' key
			nControlID = IDC_CSLASH;
			break;
		case 0xDB:	// VK_OEM_4	Used for miscellaneous characters; it can vary by keyboard. 
					//			Windows 2000/XP: For the US standard keyboard, the '[{' key
			nControlID = IDC_SQUARE_BRA;
			break;
		case 0xDC:	// VK_OEM_5	Used for miscellaneous characters; it can vary by keyboard. 
					//			Windows 2000/XP: For the US standard keyboard, the '\|' key
			nControlID = IDC_CBACKSLASH;
			break;
		case 0xE2:	// VK_OEM_102 Windows 2000/XP:	Either the angle bracket key or the backslash key
					//								on the RT 102-key keyboard
			nControlID = IDC_CBACKSLASH_106;
			break;
		default:
			break;
		}
		break;

	case CONTROL_SHIFT:
		switch (nKey) {
		case VK_RETURN:
			nControlID = IDC_CENTER;
			break;
		case 0xBD:	// VK_OEM_MINUS Windows 2000/XP: For any country/region, the '-' key
					// Shift + '-' = '_' on 101 keyboard
			nControlID = IDC_C_;
			break;
		case 0xE2:	// VK_OEM_102 Windows 2000/XP:	Either the angle bracket key or the backslash key
					//								on the RT 102-key keyboard
			nControlID = IDC_C_;
			break;
		default:
			break;
		}
		break;

	case CONTROLX:
		switch (nKey) {
		case 'C':
			nControlID = IDC_CXCC;
			break;
		case 'F':
			nControlID = IDC_CXCF;
			break;
		case 'J':
			nControlID = IDC_CXCJ;
			break;
		case 'P':
			nControlID = IDC_CXCP;
			break;
		case 'S':
			nControlID = IDC_CXCS;
			break;
		case 'W':
			nControlID = IDC_CXCW;
			break;
		default:
			break;
		}
		break;

	case CONTROL_META_SHIFT:
		switch (nKey) {
		case '5':
			nControlID = IDC_CMS5;
			break;
		default:
			break;
		}
		break;

	case META:
		switch (nKey) {
		case 'B':
			nControlID = IDC_MB;
			break;
		case 'F':
			nControlID = IDC_MF;
			break;
		case 'I':
			nControlID = IDC_MI;
			break;
		case 'V':
			nControlID = IDC_MV;
			break;
		case 'W':
			nControlID = IDC_MW;
			break;
		case 'Y':
			nControlID = IDC_MY;
			break;
		default:
			break;
		}
		break;

	case META_SHIFT:
		switch (nKey) {
		case '1':
			nControlID = IDC_MS1;
			break;
		case '5':
			nControlID = IDC_MS5;
			break;
		case 0xBC:	// VK_OEM_COMMA Windows 2000/XP: For any country/region, the ',' key
			nControlID = IDC_MSCOMMA;
			break;
		case 0xBE:	// VK_OEM_PERIOD Windows 2000/XP: For any country/region, the '.' key
			nControlID = IDC_MSPERIOD;
			break;
		default:
			break;
		}
		break;

	default:
		ASSERT(0);
		break;
	}

	return nControlID;
}

static int IsNotSameString(CComboBox* pApplication, CString szListItem);

// initiate application name combo box
void CXkeymacsDlg::InitApplicationList()
{
	m_cApplication.ResetContent();

	m_dwTasks = GetTaskList(m_TaskList, MAX_TASKS);

	EnumWindows(EnumWndProc, (LPARAM)&m_cApplication);

	for (int i = 0; i < MAX_APP; i++) {
		CString szDefault;
		szDefault.LoadString(IDS_DEFAULT);
		CString szApplicationName	= m_XkeymacsData[i].GetApplicationName();
		CString szApplicationTitle	= m_XkeymacsData[i].GetApplicationTitle();

		CString szListItem;
		szListItem.Format(IDS_APPLICATION_LIST_ITEM, szApplicationTitle, szApplicationName);
		if (IsNotSameString(&m_cApplication, szListItem) &&
			strcmp(szApplicationName, szDefault) &&
			!szApplicationName.IsEmpty()) {
			m_cApplication.AddString(szListItem);
		}
	}

	// Add IME
	HKL hKL = GetKeyboardLayout(0);
	if (ImmIsIME(hKL)) {
		char szDummy;

		UINT nIMEDescription = ImmGetDescription(hKL, &szDummy, 0);
		nIMEDescription += sizeof(char);	// for NULL
		char *szIMEDescription = new char[nIMEDescription];
		ImmGetDescription(hKL, szIMEDescription, nIMEDescription);

		UINT nIMEFileName = ImmGetIMEFileName(hKL, &szDummy, 0);
		nIMEFileName += sizeof(char);
		char *szIMEFileName = new char[nIMEFileName];
		ImmGetIMEFileName(hKL, szIMEFileName, nIMEFileName);

		CString szIMETitle;
		szIMETitle.Format(IDS_APPLICATION_LIST_ITEM, szIMEDescription, szIMEFileName);
		if (IsNotSameString(&m_cApplication, szIMETitle) &&
			(nIMEDescription || nIMEFileName)) {
			m_cApplication.AddString(szIMETitle);
		}

		delete[] szIMEDescription;
		delete[] szIMEFileName;
	}

	// Add Default
	CString szDefaultTitle;
	szDefaultTitle.LoadString(IDS_DEFAULT_TITLE);
	m_cApplication.InsertString( 0, szDefaultTitle);
	m_cApplication.SelectString(-1, szDefaultTitle);
}

static int CountSeparator(CString szMainString, CString szSeparator)
{
	int index	= 0;
	int counter	= 0;

	while ((index = szMainString.Find(szSeparator, index)) != -1) {
		index++;
		counter++;
	}

	return counter;
}

static void GetNthString(CString *szAppName, CString szWindowName,
						 CString szSeparator, int n)
{
	int index = -1;

	while (--n) {
		index = szWindowName.Find(szSeparator, index + 1);
	}

	int nStart;
	if (index != -1) {
		nStart = index + szSeparator.GetLength();
	} else {
		nStart = 0;
	}

	int nEnd = szWindowName.Find(szSeparator, nStart);
	if (nEnd == -1) {
		nEnd = szWindowName.GetLength();
	}

	*szAppName = szWindowName.Mid(nStart, nEnd - nStart);
}

static void GetAppName(CString *szAppName, char* pWindowName)
{
	CString szWindowName(pWindowName);
	CString szSeparator;
	szSeparator.LoadString(IDS_SEPARATE_WINDOWTITLE);
	int nWord = CountSeparator(szWindowName, szSeparator) + 1;

	while (nWord) {
		GetNthString(szAppName, szWindowName, szSeparator, nWord);
		if (szAppName->GetAt(0) == '['		||
			szAppName->Find('.', 0) != -1	||	// for Microsoft Project
			szAppName->Find(']', 0) != -1) {	// for the file name like [foo - bar]
			nWord--;
		} else {
			return;
		}
	}

	*szAppName = szWindowName;
}

static void Item2AppName(CString *sz)
{
	CString szDefaultTitle;
	szDefaultTitle.LoadString(IDS_DEFAULT_TITLE);
	if (*sz == szDefaultTitle) {
		sz->LoadString(IDS_DEFAULT);
	}

	int nStart, nEnd, nCount;

	nStart	= sz->ReverseFind('(') + 1;
	nEnd	= sz->Find(')', nStart) - 1;
	nCount	= (nEnd + 1) - nStart;
	*sz		= sz->Mid(nStart, nCount);
}

static int IsNotSameString(CComboBox* pApplication, CString szListItem)
{
	CString szItem, szList;
	szList = szListItem;
	Item2AppName(&szList);

	for (int i = 0; i < pApplication->GetCount(); i++) {
		pApplication->GetLBText(i, szItem);
		Item2AppName(&szItem);
		if (!stricmp(szItem, szList)) {
			return 0;
		}
	}

	return 1;
}

BOOL CALLBACK EnumWndProc(HWND hWnd, LPARAM lParam)
{
	CComboBox		*pApplication	= (CComboBox*)lParam;
	CXkeymacsDlg	*pXkeymacsDlg	= (CXkeymacsDlg*)pApplication->GetParent();
	PTASK_LIST		pTask			= pXkeymacsDlg->m_TaskList;
	
	char szWindowName[WINDOW_NAME_LENGTH];
	char szClassName[CLASS_NAME_LENGTH];
	WINDOWPLACEMENT wpl;
	
	wpl.length = sizeof(WINDOWPLACEMENT);
	GetWindowText(hWnd, szWindowName, sizeof(szWindowName));
	GetClassName(hWnd, szClassName, sizeof(szClassName));

	CString szAppName;
	// Get Process Name
	DWORD dwProcessId = 0;
	GetWindowThreadProcessId(hWnd, &dwProcessId);
	for (DWORD i = 0; i < pXkeymacsDlg->m_dwTasks; i++) {
		if (pTask[i].dwProcessId == dwProcessId) {
			CString szB2;
			szB2.LoadString(IDS_B2);
			CString szExplorer;
			szExplorer.LoadString(IDS_EXPLORER);
			CString szProject;
			szProject.LoadString(IDS_PROJECT);
			CString szExcel;
			szExcel.LoadString(IDS_EXCEL);
			CString szMSDN;
			szMSDN.LoadString(IDS_MSDN);
			CString szMsimn;
			szMsimn.LoadString(IDS_MSIMN);
			
			// Get Application Name
			if (szWindowName[0] == '\0') {
				continue;
			}
			if (strnicmp(pTask[i].ProcessName, szB2, sizeof(pTask[i].ProcessName)) &&
				strnicmp(pTask[i].ProcessName, szExplorer, sizeof(pTask[i].ProcessName)) &&
				strnicmp(pTask[i].ProcessName, szProject, sizeof(pTask[i].ProcessName)) &&
				strnicmp(pTask[i].ProcessName, szExcel, sizeof(pTask[i].ProcessName)) &&
				strnicmp(pTask[i].ProcessName, szMSDN, sizeof(pTask[i].ProcessName)) &&
				strnicmp(pTask[i].ProcessName, szMsimn, sizeof(pTask[i].ProcessName))) {
				GetAppName(&szAppName, szWindowName);
			} else if (!strnicmp(pTask[i].ProcessName, szB2, sizeof(pTask[i].ProcessName))) {
				szAppName.LoadString(IDS_BECKY);
			} else if (!strnicmp(pTask[i].ProcessName, szExplorer, sizeof(pTask[i].ProcessName))) {
				szAppName.LoadString(IDS_PROGRAM_MANAGER);
			} else if (!strnicmp(pTask[i].ProcessName, szMsimn, sizeof(pTask[i].ProcessName))) {
				szAppName.LoadString(IDS_OUTLOOK_EXPRESS);
			} else if (!strnicmp(pTask[i].ProcessName, szProject, sizeof(pTask[i].ProcessName)) ||
					 !strnicmp(pTask[i].ProcessName, szExcel, sizeof(pTask[i].ProcessName))) {
				CString szSeparator;
				szSeparator.LoadString(IDS_SEPARATE_WINDOWTITLE);
				GetNthString(&szAppName, szWindowName, szSeparator, 1);
			} else if (!strnicmp(pTask[i].ProcessName, szMSDN, sizeof(pTask[i].ProcessName))) {
				szAppName = szWindowName;
			} else {
				szAppName.LoadString(IDS_UNKNOWN_APPLICATION);
			}
			break;
		}
	}
	
	
	if ((IsWindowVisible(hWnd)) &&								// Is visible?
		(GetWindow(hWnd, GW_OWNER) == NULL) &&					// Is top level window?
		(lstrlen(szWindowName) > 0) &&							// Have caption?
		(pApplication->FindString(-1, szClassName) == CB_ERR)) {// Is not same string?
		CString szListItem;
//		szListItem.Format(IDS_APPLICATION_LIST_ITEM, szClassName, szWindowName);		
		szListItem.Format(IDS_APPLICATION_LIST_ITEM, szAppName, pTask[i].ProcessName);
		if (IsNotSameString(pApplication, szListItem)) {
			pApplication->AddString(szListItem);
		}
	}
	return TRUE;
}

void CXkeymacsDlg::OnSelchangeApplication() 
{
	// TODO: Add your control notification handler code here
	GetDialogData(m_szCurrentApplication);

	GetApplicationTitle(m_szCurrentApplication, m_cApplication.GetCurSel());
	int nStart;
	if ((nStart = m_szCurrentApplication.ReverseFind('(')) != -1) {
		nStart++;
		m_szCurrentApplication =
			m_szCurrentApplication.Mid(nStart, m_szCurrentApplication.GetLength() - nStart - 1);
	}

	GetDlgItem(IDC_LOAD_DEFAULT)->EnableWindow(TRUE);
	SetDialogData(m_szCurrentApplication);
	if (IsDefault(m_szCurrentApplication)) {
		GetDlgItem(IDC_LOAD_DEFAULT)->EnableWindow(FALSE);
	}
}

void CXkeymacsDlg::OnClearAll() 
{
	// TODO: Add your control notification handler code here
	SetAllDialogData(0);
}

void CXkeymacsDlg::OnSelectAll() 
{
	// TODO: Add your control notification handler code here
	SetAllDialogData(1);
}

// set all dialog data
void CXkeymacsDlg::SetAllDialogData(UINT nCheck)
{
	for (int nCommandType = 0; nCommandType < MAX_COMMAND_TYPE; nCommandType++) {
		for (int nKey = 0; nKey < MAX_KEY; nKey++) {
			int nControlID = GetControlID(nCommandType, nKey);
			if (!nControlID) {
				continue;
			}
			CheckDlgButton(nControlID, nCheck);
		}
	}
	CheckDlgButton(IDC_CO2, BST_UNCHECKED);
}

void CXkeymacsDlg::OnLoadDefault() 
{
	// TODO: Add your control notification handler code here
	CString szDefault;
	szDefault.LoadString(IDS_DEFAULT);
	SetDialogData(szDefault);
	EnableUseDefaultButton(TRUE);
}

// delete all registry data
void CXkeymacsDlg::DeleteAllRegistryData(HKEY hkey)
{
	DWORD dwIndex = 0;
	char szName[SUB_KEY_NAME_LENGTH];
	DWORD dwName = sizeof(szName);
	ZeroMemory(szName, sizeof(szName));
	FILETIME filetime;

	while (RegEnumKeyEx(hkey, dwIndex, szName, &dwName, NULL, NULL, NULL, &filetime) == ERROR_SUCCESS) {
		VERIFY(RegDeleteKey(hkey, szName) == ERROR_SUCCESS);
		ZeroMemory(szName, sizeof(szName));
	}
}

void CXkeymacsDlg::OnCo()
{
	// TODO: Add your control notification handler code here
	if (IsDlgButtonChecked(IDC_CO) == BST_CHECKED) {
		CheckDlgButton(IDC_CO2, BST_UNCHECKED);
	}
}

void CXkeymacsDlg::OnCo2()
{
	// TODO: Add your control notification handler code here
	if (IsDlgButtonChecked(IDC_CO2) == BST_CHECKED) {
		CheckDlgButton(IDC_CO, BST_UNCHECKED);
	}
}

void CXkeymacsDlg::OnUseDefault() 
{
	// TODO: Add your control notification handler code here
	BOOL bUseDefault;
	if (m_cUseDefault.GetCheck()) {
		bUseDefault = TRUE;
	} else if (GetSavedSettingCount() >= MAX_APP) {
		bUseDefault = TRUE;
		m_cUseDefault.SetCheck(1);

		CString szText;
		szText.Format(IDS_ERR_SAVE_DATA, MAX_APP-1);
		MessageBox(szText);
	} else {
		bUseDefault = FALSE;
	}
	UseDefault(bUseDefault);
}

// set control enable/disable
void CXkeymacsDlg::UseDefault(BOOL bUseDefault)
{
	for (int nCommandType = 0; nCommandType < MAX_COMMAND_TYPE; nCommandType++) {
		for (int nKey = 0; nKey < MAX_KEY; nKey++) {
			int nControlID = GetControlID(nCommandType, nKey);
			if (!nControlID) {
				continue;
			}
			GetDlgItem(nControlID)->EnableWindow(!bUseDefault);
		}
	}
	GetDlgItem(IDC_CO2)->EnableWindow(!bUseDefault);
	GetDlgItem(IDC_KILL_RING_MAX)->EnableWindow(!bUseDefault);
	GetDlgItem(IDC_LOAD_DEFAULT)->EnableWindow(!bUseDefault);
	GetDlgItem(IDC_SELECT_ALL)->EnableWindow(!bUseDefault);
	GetDlgItem(IDC_CLEAR_ALL)->EnableWindow(!bUseDefault);
}

// get IDC_USE_DEFAULT is checked or not, and change control on main dialog enable/disable
void CXkeymacsDlg::CheckUseDefault(BOOL bCheckUseDefault)
{
	CheckDlgButton(IDC_USE_DEFAULT, bCheckUseDefault);
	UseDefault(bCheckUseDefault);
}

// enable Use Default Setting button
void CXkeymacsDlg::EnableUseDefaultButton(BOOL bEnable)
{
	GetDlgItem(IDC_USE_DEFAULT)->EnableWindow(bEnable);
}

// initiate dll data
void CXkeymacsDlg::InitDllData()
{
	LoadRegistryData();
	SetDllData();
}

void CXkeymacsDlg::OnDropdownApplication() 
{
	// TODO: Add your control notification handler code here
	InitApplicationList();

	int nID = GetCurrentApplicationID();
	if (nID != CB_ERR) {
		m_cApplication.SetCurSel(nID);
	} else {
		GetDialogData(m_szCurrentApplication);
		m_szCurrentApplication.LoadString(IDS_DEFAULT_TITLE);
		m_cApplication.SelectString(-1, m_szCurrentApplication);
		SetDialogData(m_szCurrentApplication);
	}
}

// if sz is "Default", return TRUE
BOOL CXkeymacsDlg::IsDefault(CString sz)
{
	CString szDefault;
	szDefault.LoadString(IDS_DEFAULT);

	if (szDefault == sz) {
		return TRUE;
	}

	return FALSE;
}

// return count of saved settings
int CXkeymacsDlg::GetSavedSettingCount()
{
	int nSavedSetting = 0;

	for (int nApp = 0; nApp < MAX_APP; nApp++) {
		CString szApplicationName;
		szApplicationName = m_XkeymacsData[nApp].GetApplicationName();
		if (!szApplicationName.IsEmpty()) {
			nSavedSetting++;
		}
	}

	return nSavedSetting;
}

void CXkeymacsDlg::SetDllData()
{
	CMainFrame* pMainFrame = (CMainFrame*)AfxGetMainWnd();

	for (int nApp = 0; nApp < MAX_APP; nApp++) {

		CString szApplicationName = m_XkeymacsData[nApp].GetApplicationName();

		if (szApplicationName.IsEmpty()) {
			pMainFrame->m_pXkeymacsDll->Clear(nApp);
			continue;
		}

		pMainFrame->m_pXkeymacsDll->SetApplicationName(nApp, szApplicationName);
		pMainFrame->m_pXkeymacsDll->SetEnable(nApp, CONTROL, 'X', 0);

		for (int nCommandType = 0; nCommandType < MAX_COMMAND_TYPE; nCommandType++) {
			for (int nKey = 0; nKey < MAX_KEY; nKey++) {
				if (!GetControlID(nCommandType, nKey)) {
					continue;
				}

				int nEnable = m_XkeymacsData[nApp].GetEnable(nCommandType, nKey);

				pMainFrame->m_pXkeymacsDll->SetEnable(nApp, nCommandType, nKey, nEnable);
				if (nCommandType == CONTROLX && nEnable) {
					pMainFrame->m_pXkeymacsDll->SetEnable(nApp, CONTROL, 'X', 1);
				}
			}
		}

		pMainFrame->m_pXkeymacsDll->SetKillRingMax(nApp, m_XkeymacsData[nApp].GetKillRingMax());
	}
}

#include <winperf.h>   // for Windows NT
// Add running application's names to the list
// cf. tlist of a sample program written by Microsoft that says
/*++

Routine Description:

    Provides an API for getting a list of tasks running at the time of the
    API call.  This function uses the registry performance data to get the
    task list and is therefore straight WIN32 calls that anyone can call.

Arguments:

    dwNumTasks       - maximum number of tasks that the pTask array can hold

Return Value:

    Number of tasks placed into the pTask array.

--*/
DWORD CXkeymacsDlg::GetTaskListNT(PTASK_LIST pTask, DWORD dwNumTasks)
{
	DWORD	rc;
	DWORD	i;
	HKEY	hKeyNames;
	DWORD	dwType;
	DWORD	dwSize;
	LPBYTE	buf = NULL;
	CHAR	szSubKey[1024];

	//
	// Look for the list of counters.  Always use the neutral
	// English version, regardless of the local language.  We
	// are looking for some particular keys, and we are always
	// going to do our looking in English.  We are not going
	// to show the user the counter names, so there is no need
	// to go find the corresponding name in the local language.
	//
	{
		LANGID	lid = MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL);
		CString szRegsubkeyPere;
		szRegsubkeyPere.LoadString(IDS_REGSUBKEY_PERF);
		sprintf(szSubKey, szRegsubkeyPere, lid);
		rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, szSubKey, 0, KEY_READ, &hKeyNames);
	}
	if (rc != ERROR_SUCCESS) {
		goto exit;
	}

	//
	// get the buffer size for the counter names
	//
	LoadString(AfxGetInstanceHandle(), IDS_REGSUBKEY_COUNTERS, szSubKey, sizeof(szSubKey));
	rc = RegQueryValueEx(hKeyNames, szSubKey, NULL, &dwType, NULL, &dwSize);

	if (rc != ERROR_SUCCESS) {
		goto exit;
	}

	//
	// allocate the counter names buffer
	//
	buf = (LPBYTE) calloc(dwSize, sizeof(BYTE));
	if (buf == NULL) {
		goto exit;
	}

	//
	// read the counter names from the registry
	//
	LoadString(AfxGetInstanceHandle(), IDS_REGSUBKEY_COUNTERS, szSubKey, sizeof(szSubKey));
	rc = RegQueryValueEx(hKeyNames, szSubKey, NULL, &dwType, buf, &dwSize);

	if (rc != ERROR_SUCCESS) {
		goto exit;
	}

	//
	// now loop thru the counter names looking for the following counters:
	//
	//      1.  "Process"           process name
	//      2.  "ID Process"        process id
	//
	// the buffer contains multiple null terminated strings and then
	// finally null terminated at the end.  the strings are in pairs of
	// counter number and counter name.
	//
	LPSTR p;
	LPSTR p2;
	DWORD dwProcessIdTitle;
	p = (LPSTR)buf;
	while (*p) {
		if (p > (LPSTR)buf) {
			for (p2 = p - 2; isdigit(*p2); p2--) {
				;
			}
		}

		CString szProcess;
		szProcess.LoadString(IDS_PROCESS);
		CString szProcessId;
		szProcessId.LoadString(IDS_PROCESSID);

		if (stricmp(p, szProcess) == 0) {
			//
			// look backwards for the counter number
			//
			for (p2 = p - 2; isdigit(*p2); p2--) {
				;
			}
			strcpy( szSubKey, p2 + 1 );
		} else if (stricmp(p, szProcessId) == 0) {
			//
			// look backwards for the counter number
			//
			for (p2 = p - 2; isdigit(*p2); p2--) {
				;
			}
			dwProcessIdTitle = atol(p2 + 1);
		}
		//
		// next string
		//
		p += (strlen(p) + 1);
	}

	//
	// free the counter names buffer
	//
	free(buf);


	//
	// allocate the initial buffer for the performance data
	//
	dwSize = INITIAL_SIZE;
	buf = (LPBYTE) calloc(dwSize, sizeof(BYTE));
	if (buf == NULL) {
		goto exit;
	}

	PPERF_DATA_BLOCK pPerf;
	while (TRUE) {

		rc = RegQueryValueEx( HKEY_PERFORMANCE_DATA, szSubKey, NULL, &dwType, buf, &dwSize);

		pPerf = (PPERF_DATA_BLOCK) buf;

		//
		// check for success and valid perf data block signature
		//
		if ((rc == ERROR_SUCCESS)				&&
			(dwSize > 0)						&&
			(pPerf)->Signature[0] == (WCHAR)'P' &&
			(pPerf)->Signature[1] == (WCHAR)'E' &&
			(pPerf)->Signature[2] == (WCHAR)'R' &&
			(pPerf)->Signature[3] == (WCHAR)'F' ) {
			break;
		}

		//
		// if buffer is not big enough, reallocate and try again
		//
		if (rc == ERROR_MORE_DATA) {
			dwSize += EXTEND_SIZE;
			buf = (LPBYTE) realloc( buf, dwSize );
			memset( buf, 0, dwSize );
		} else {
			goto exit;
		}
	}

	//
	// set the perf_object_type pointer
	//
	PPERF_OBJECT_TYPE pObj;
	pObj = (PPERF_OBJECT_TYPE) ((DWORD)pPerf + pPerf->HeaderLength);

	//
	// loop thru the performance counter definition records looking
	// for the process id counter and then save its offset
	//
	DWORD dwProcessIdCounter;
	PPERF_COUNTER_DEFINITION pCounterDef;
	pCounterDef = (PPERF_COUNTER_DEFINITION) ((DWORD)pObj + pObj->HeaderLength);
	for (i = 0; i < pObj->NumCounters; i++) {
		if (pCounterDef->CounterNameTitleIndex == dwProcessIdTitle) {
			dwProcessIdCounter = pCounterDef->CounterOffset;
			break;
		}
		pCounterDef++;
	}

	DWORD dwLimit;
	dwLimit = dwNumTasks - 1;
	dwNumTasks = min( dwLimit, (DWORD)pObj->NumInstances );

	PPERF_INSTANCE_DEFINITION pInst;
	pInst = (PPERF_INSTANCE_DEFINITION) ((DWORD)pObj + pObj->DefinitionLength);

	//
	// loop thru the performance instance data extracting each process name
	// and process id
	//
	for (i = 0; i < dwNumTasks; i++) {
		//
		// pointer to the process name
		//
		p = (LPSTR) ((DWORD)pInst + pInst->NameOffset);

		//
		// convert it to ascii
		//
		CHAR szProcessName[MAX_PATH];
		rc = WideCharToMultiByte(CP_ACP, 0, (LPCWSTR)p, -1, szProcessName,
								 sizeof(szProcessName), NULL, NULL);

		if (!rc) {
			//
			// if we cant convert the string then use a default value
			//
			CString szUnknownTask;
			szUnknownTask.LoadString(IDS_UNKNOWN_TASK);
			strcpy(pTask->ProcessName, szUnknownTask);
		}

		if (strlen(szProcessName)+4 <= sizeof(pTask->ProcessName)) {
			CString szExe;
			szExe.LoadString(IDS_EXTENSION_EXECUTABLE);

			strcpy(pTask->ProcessName, szProcessName);
			strcat(pTask->ProcessName, szExe);
		}

		//
		// get the process id
		//
		PPERF_COUNTER_BLOCK pCounter = (PPERF_COUNTER_BLOCK) ((DWORD)pInst + pInst->ByteLength);
//		pTask->flags = 0;
		pTask->dwProcessId = *((LPDWORD) ((DWORD)pCounter + dwProcessIdCounter));
		if (pTask->dwProcessId == 0) {
			pTask->dwProcessId = (DWORD) - 2;
		}

		//
		// next process
		//
		pTask++;
		pInst = (PPERF_INSTANCE_DEFINITION) ((DWORD)pCounter + pCounter->ByteLength);
	}

exit:
	if (buf) {
		free(buf);
	}

	RegCloseKey( hKeyNames );
	RegCloseKey( HKEY_PERFORMANCE_DATA );

	return dwNumTasks;
//	return;
}


/*++

Routine Description:

    Provides an API for getting a list of tasks running at the time of the
    API call.  This function uses Toolhelp32to get the task list and is 
    therefore straight WIN32 calls that anyone can call.

Arguments:

    dwNumTasks       - maximum number of tasks that the pTask array can hold

Return Value:

    Number of tasks placed into the pTask array.

--*/
DWORD CXkeymacsDlg::GetTaskList95(PTASK_LIST pTask, DWORD dwNumTasks)
{
	CREATESNAPSHOT pCreateToolhelp32Snapshot = NULL;
	PROCESSWALK    pProcess32First           = NULL;
	PROCESSWALK    pProcess32Next            = NULL;

	HMODULE        hKernel        = NULL;
	HANDLE         hProcessSnap   = NULL;
	PROCESSENTRY32 pe32           = {0};
	DWORD          dwTaskCount    = 0;


	// Guarantee to the code later on that we'll enum at least one task.
	if (dwNumTasks == 0) {
		return 0;
	}

	// Obtain a module handle to KERNEL so that we can get the addresses of
	// the 32-bit Toolhelp functions we need.
	hKernel = GetModuleHandle("KERNEL32.DLL");

	if (hKernel) {
		pCreateToolhelp32Snapshot =
			(CREATESNAPSHOT)GetProcAddress(hKernel, "CreateToolhelp32Snapshot");

		pProcess32First = (PROCESSWALK)GetProcAddress(hKernel, "Process32First");
		pProcess32Next  = (PROCESSWALK)GetProcAddress(hKernel, "Process32Next");
	}

	// make sure we got addresses of all needed Toolhelp functions.
	if (!(pProcess32First && pProcess32Next && pCreateToolhelp32Snapshot)) {
		return 0;
	}

	// Take a snapshot of all processes currently in the system.
	hProcessSnap = pCreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
	if (hProcessSnap == (HANDLE)-1) {
		return 0;
	}

	// Walk the snapshot of processes and for each process, get information
	// to display.
	dwTaskCount = 0;
	pe32.dwSize = sizeof(PROCESSENTRY32);   // must be filled out before use
	if (pProcess32First(hProcessSnap, &pe32)) {
		do {
			LPSTR pCurChar;

			// strip path and leave executabe filename splitpath
			for (pCurChar = (pe32.szExeFile + lstrlen (pe32.szExeFile));
				*pCurChar != '\\' && pCurChar != pe32.szExeFile; 
				--pCurChar)

			lstrcpy(pTask -> ProcessName, pCurChar);
//			pTask -> flags = 0;
			pTask -> dwProcessId = pe32.th32ProcessID;

			++dwTaskCount;   // keep track of how many tasks we've got so far
			++pTask;         // get to next task info block.
		} while (dwTaskCount < dwNumTasks && pProcess32Next(hProcessSnap, &pe32));
	} else {
		dwTaskCount = 0;    // Couldn't walk the list of processes.
	}

	// Don't forget to clean up the snapshot object...
	CloseHandle (hProcessSnap);

	return dwTaskCount;
}

DWORD CXkeymacsDlg::GetTaskList(PTASK_LIST pTask, DWORD dwNumTasks)
{

for (int i = 0; i < MAX_TASKS; i++) {
	ZeroMemory(&pTask[i], sizeof(PTASK_LIST));
//	ZeroMemory(pTask[i].ProcessName, sizeof(pTask[i].ProcessName));
//	pTask[i].dwProcessId = 0;
}

	OSVERSIONINFO     verInfo = {0};

	verInfo.dwOSVersionInfoSize = sizeof (verInfo);
	GetVersionEx(&verInfo);

	switch (verInfo.dwPlatformId) {
	case VER_PLATFORM_WIN32_NT:
		return GetTaskListNT(pTask, dwNumTasks);
		break;

	case VER_PLATFORM_WIN32_WINDOWS:
		return GetTaskList95(pTask, dwNumTasks);
		break;

	default:
		AfxMessageBox(IDS_ERR_WINDOWS_VERSION);
		return 0;
	}
}


int CXkeymacsDlg::GetCurrentApplicationID()
{
	int nCounter = m_cApplication.GetCount();
	CString szListItem;
	int nCurSel = m_cApplication.GetCurSel();

	for (int i = 0; i < nCounter; i++) {
		m_cApplication.SetCurSel(i);
		GetApplicationTitle(szListItem);
		if (szListItem.Find(m_szCurrentApplication) != -1) {
			m_cApplication.SetCurSel(nCurSel);
			return i;
		}
	}
	return -1;
}

void CXkeymacsDlg::GetApplicationTitle(CString &rList, int nIndex)
{
	if (0 <= nIndex) {
		m_cApplication.GetLBText(nIndex, rList);
	} else {
		m_cApplication.GetWindowText(rList);
	}

	CString szDefaultTitle;
	szDefaultTitle.LoadString(IDS_DEFAULT_TITLE);
	if (rList == szDefaultTitle) {
		rList.LoadString(IDS_DEFAULT);
	}
	return;
}
